package org.springblade.sms.component;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.springblade.sms.bean.params.TemplateApplyResponse;
import org.springblade.sms.bean.params.TemplateCheckResponse;
import org.springblade.useranalysis.common.utils.DesUtil;
import org.springblade.useranalysis.common.utils.HashKit;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONObject;
import com.github.xiaoymin.knife4j.core.util.StrUtil;

import cn.hutool.crypto.digest.MD5;
import cn.hutool.http.HttpUtil;
import lombok.extern.slf4j.Slf4j;

/**
 * 
 * 短信通用操作，支持传递短信模板ID
 * 
 * 短信通道：该渠道使用10655117网关号发送，队列没有限制，最大支持数据100条/秒，70个字符为一条
 * 
 * @author william
 * @date 2022年8月23日 下午2:30:39
 */
@Slf4j
@Component
public class SmsSender {
	final static String BASE_URL = "http://%s.mige.tv/smsbusi";
	final String DOMAIN = "cctv2";

	@Value("${spring.profiles.active:prod}")
	private String activeProfile;

	public String host() {
		return String.format(BASE_URL, DOMAIN);
	}

	public boolean isProd() {
		return activeProfile != null && activeProfile.equalsIgnoreCase("prod");
	}


	/**
	 * 短信发送，支持彩信，具体和短信模板ID（contentid有关）
	 * 
	 * @param varMap 模板参数，不同的模板有不同的参数，格式为json字符串，示例：{name:tom,age:20}
	 * @param mobile 手机号
	 * @param contentId 模板ID
	 * @param smsConfig 短信配置
	 * @return
	 */
	public JSONObject sendSms(Map<String, String> varMap, String mobile, String contentId, SmsConfig smsConfig) {
		try {
			log.info("sendSms:smsid is " + smsConfig.getSmsid());
			log.info("sendSms:pwk is " + smsConfig.getPwd());
			log.info("sendSms:contentId is " + contentId);
			log.info("sendSms:varMap is " + JSON.toJSONString(varMap));
			log.info("sendSms:mobile is " + mobile);

			String userid = DesUtil.encode(mobile, smsConfig.getPwd());
			log.info("sendSms:userid is " + userid);
			String timestamp = String.valueOf(System.currentTimeMillis() / 1000);
			String vars = JSON.toJSONString(varMap);
			Map<String, Object> requestMap = new HashMap<String, Object>();
			requestMap.put("smsid", smsConfig.getSmsid());
			requestMap.put("userid", userid);
			requestMap.put("timestamp", timestamp);
			requestMap.put("vars", vars);
			requestMap.put("contentid", contentId);
			requestMap.put("content", "");
			List<String> signList = Arrays.asList(smsConfig.getSmsid(), userid, timestamp, vars, contentId);
			String sign = HashKit.sha1(String.join("", signList) + smsConfig.getPwd()).toLowerCase();
			requestMap.put("sign", sign);
			// 新短信网关接口需要，1-验证码；2-营销短信；3-其它短信
			requestMap.put("type", 2);
			String response = HttpUtil.get(host() + "/sms/send", requestMap);
			log.info(String.format("SMS:send request|mobile=%s,response=%s", mobile, response));
			// {resultcode: "0", errorinfo: "", taskid: ""}
			if (response != null && !"".equals(response)) {
				JSONObject resultObject = JSON.parseObject(response);
				return resultObject;
			}
		} catch (Exception e) {
			log.error(e.getMessage(), e);
		}

		return null;
	}

	public int querySmsStatus(String mobile, String smsBizId, SmsConfig smsConfig) {
		try {
			String userid = DesUtil.encode(mobile, smsConfig.getPwd());
			String timestamp = String.valueOf(System.currentTimeMillis() / 1000);
			Map<String, Object> requestMap = new HashMap<String, Object>();
			requestMap.put("smsid", smsConfig.getSmsid());
			requestMap.put("userid", userid);
			requestMap.put("timestamp", timestamp);
			requestMap.put("taskid", smsBizId);
			List<String> signList = Arrays.asList(smsConfig.getSmsid(), userid, timestamp, smsBizId);
			String sign = HashKit.sha1(String.join("", signList) + smsConfig.getPwd()).toLowerCase();
			requestMap.put("sign", sign);
			String response = HttpUtil.get(host() + "/sms/query", requestMap);
			log.info(String.format("SmsSender:querySmsStatus|mobile=%s,response=%s", mobile, response));
			// {resultcode: "0", errorinfo: "", taskid: "",status:"0"}
			// status【0/1/2/3/5】，分别对应状态：（已发送/发送成功/未知结果/点击/失败）
			// 0 已发送
			// 1 发送成功
			// 2 未知结果
			// 3 点击
			// 5 失败
			if (response != null && !"".equals(response)) {
				JSONObject resultObject = JSON.parseObject(response);
				int status = resultObject.getIntValue("status");
				return status;
			}
		} catch (Exception e) {
			log.error("SmsSender:querySmsStatus error is: {}" + e.getMessage(), e);
		}

		return 2;
	}

	/**
	 * 数字短信模版上传
	 * 
	 * @param title 标题
	 * @param content 内容
	 * @param smsid 渠道ID
	 * @param smsConfig 短信配置信息
	 * @return
	 */
	public TemplateApplyResponse updtempsms(String title, String content, SmsConfig smsConfig) {
		try {
			String timestamp = String.valueOf(System.currentTimeMillis() / 1000);
			Map<String, Object> requestMap = new HashMap<String, Object>();
			requestMap.put("title", title);
			requestMap.put("content", content);
			requestMap.put("timestamp", timestamp);
			requestMap.put("smsid", smsConfig.getSmsid());
			List<String> signList = Arrays.asList(smsConfig.getSmsid(), timestamp);
			String sign = MD5.create().digestHex(String.join("", signList) + smsConfig.getPwd()).toLowerCase();
			requestMap.put("signature", sign);
			String json = JSON.toJSONString(requestMap);
			log.info("json is " + json);
			String response = HttpUtil.post(host() + "/digitalsms/updtempsms", json);
			log.info(String.format("SMS:template apply request|response=%s", response));
			if (StrUtil.isNotBlank(response)) {
				TemplateApplyResponse tar = null;
				try {
					tar = JSON.parseObject(response, TemplateApplyResponse.class);
					tar.setResponse(response);
				} catch (Exception e) {
					tar = new TemplateApplyResponse(response);
				}
				return tar;
			}
		} catch (Exception e) {
			log.error(e.getMessage(), e);
		}

		return null;
	}

	/**
	 * 查询聚合数字短信模板审核状态
	 * 
	 * @param mmsid
	 */
	public TemplateCheckResponse qrytempstatus(String mmsid, SmsConfig smsConfig) {
		try {
			String timestamp = String.valueOf(System.currentTimeMillis() / 1000);
			Map<String, Object> requestMap = new HashMap<String, Object>();
			requestMap.put("mmsid", mmsid);
			requestMap.put("timestamp", timestamp);
			requestMap.put("smsid", smsConfig.getSmsid());
			List<String> signList = Arrays.asList(smsConfig.getSmsid(), timestamp);
			String sign = MD5.create().digestHex(String.join("", signList) + smsConfig.getPwd()).toLowerCase();
			requestMap.put("signature", sign);
			String json = JSON.toJSONString(requestMap);
			String response = HttpUtil.post(host() + "/digitalsms/qrytempstatus", json);
			log.info(String.format("SMS:template check request|response=%s", response));
			if (StrUtil.isNotBlank(response)) {
				TemplateCheckResponse tar = JSON.parseObject(response, TemplateCheckResponse.class);
				return tar;
			}
		} catch (Exception e) {
			log.error(e.getMessage(), e);
		}

		return null;
	}

	public static void main(String[] args) throws Exception {
//		SmsSender sender = new SmsSender();
//		log.info("prod is " + sender.isProd());
//
//		String title = "测试数字短信上传";
//		String content = "你好${p1}，测试一下";
//		String base64 = Base64.encode(content.getBytes(Charset.forName("gb2312")));
//		sender.updtempsms(title, ",txt|" + base64, SmsConfigBase.of(SmsType.digital));
//		sender.qrytempstatus("20220830112324473467482", SmsConfigBase.of(SmsType.digital));
//		

//		String json = "{\"resultcode\":null,\"resultInfo\":null,\"data\":\"ip地址校验失败(39.99.245.43)\"}";
//		JSONObject jo = JSON.parseObject(json);
//		String errorInfo = jo.getString("data");
//		log.info(""+errorInfo);
	}
}
