package org.springblade.useranalysis.common.utils;

import java.io.Serializable;
import java.lang.reflect.ParameterizedType;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;

import com.alibaba.excel.context.AnalysisContext;
import com.alibaba.excel.event.AnalysisEventListener;
import com.alibaba.fastjson2.JSON;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;

import lombok.extern.slf4j.Slf4j;

/**
 * 通用型分页导入excel
 * 
 * @author Administrator
 *
 * @param <T>
 * @param <E>
 * @param <S>
 */
@Slf4j
public abstract class CommonExcelListener<T, E extends Serializable, S extends ServiceImpl<?, T>>
		extends AnalysisEventListener<E> {

	/**
	 * 每隔500条存储数据库，实际使用中可以3000条，然后清理list ，方便内存回收
	 */
	private static final int BATCH_COUNT = 3000;
	List<E> list = new ArrayList<E>();

	private S bizService;
	private Function<E, T> function;

	private long total;

	/**
	 * 如果使用了spring,请使用这个构造方法。每次创建Listener的时候需要把spring管理的类传进来
	 * 
	 * @param bizService
	 */
	public CommonExcelListener(S bizService) {
		this.bizService = bizService;
	}

	public CommonExcelListener(S bizService, Function<E, T> function) {
		this.function = function;
		this.bizService = bizService;
	}

	/**
	 * 每一条数据解析都会来调用
	 */
	@Override
	public void invoke(E data, AnalysisContext context) {
		log.info("解析到一条数据:{}", JSON.toJSONString(data));
		list.add(data);
		// 达到BATCH_COUNT了，需要去存储一次数据库，防止数据几万条数据在内存，容易OOM
		if (list.size() >= BATCH_COUNT) {
			saveData();
			// 存储完成清理 list
			list.clear();
		}
	}

	/**
	 * 所有数据解析完成了 都会来调用
	 */
	@Override
	public void doAfterAllAnalysed(AnalysisContext context) {
		saveData();
		log.info("所有数据解析完成！");
	}

	@SuppressWarnings("unchecked")
	private Class<T> getTClass() {
		Class<T> clz = (Class<T>) ((ParameterizedType) getClass().getGenericSuperclass()).getActualTypeArguments()[0];
		return clz;
	}

	/**
	 * 将数据存储数据库
	 */
	private void saveData() {
		log.info("{}条数据，开始存储数据库！", list.size());
		List<T> modelList = new ArrayList<T>();
		list.stream().forEach(f -> {
			if (function != null) {
				modelList.add(function.apply(f));
			} else {
				modelList.add(ConvertKit.transfer(f, getTClass()));
			}
		});
		bizService.saveBatch(modelList);
		total += modelList.size();
		log.info("存储数据库成功！");
	}

	public long getTotal() {
		return total;
	}

}
