/**
 * 
 */
package org.springblade.useranalysis.service.normal;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.stream.Collectors;

import org.springblade.useranalysis.entity.AdsUserprofileTag;
import org.springblade.useranalysis.entity.UserGroupAsyncDataLog;
import org.springblade.useranalysis.entity.UserGroupEntity;
import org.springblade.useranalysis.entity.clickhouse.ZUserTagEntity;
import org.springblade.useranalysis.enums.UserGroupProcessStatus;
import org.springblade.useranalysis.service.UserGroupService;
import org.springblade.useranalysis.service.UserProfileService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.CommandLineRunner;
import org.springframework.stereotype.Service;

import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;

import cn.hutool.core.date.StopWatch;
import cn.hutool.core.util.StrUtil;
import lombok.extern.slf4j.Slf4j;

/**
 * @author wookvn
 *
 * 仅有一个环境可以执行
 * --blade.task.sumdata=true
 */
@Slf4j
@Service
public class UserGroupSumDataManager implements CommandLineRunner {
	
	private final LinkedBlockingQueue<String> queue = new LinkedBlockingQueue<>();
	
	@Autowired
	private UserGroupService userGroupService;
	
	@Autowired
	private UserProfileService userProfileService;
	
	@Autowired
	private UserGroupUserIdService userGroupUserIdService;
	
	@Autowired
	private UserGroupRuleService userGroupRuleService;
	
	@Autowired
	private UserGroupAsyncDataLogService userGroupAsyncDataLogService;
	
	@Autowired
	private UserGroupSumDataService userGroupSumDataService;
	
	@Value("${blade.task.sumdata:false}")
	private boolean flag;
	
	@Value("${spring.profiles.active:prod}")
	private String activeProfile;
	
	/**
	 * 计算需要处理哪些月份的数据，往队列推数据等待执行
	 * @param groupId
	 */
	public boolean execute(String groupId) {
		try {
			//判断队列里面不存在
			if(!queue.stream().anyMatch(item -> item.equals(groupId))) {
				queue.put(groupId);
				//更新人群处理状态为：队列状态
				UserGroupEntity entity = new UserGroupEntity();
				entity.setGroupId(groupId);
				entity.setProcessStatus(UserGroupProcessStatus.queueing);
				userGroupService.updateById(entity);
				return true;
			}
		} catch (Exception e) {
			log.error(e.getMessage());
		}
		return false;
	}

	/**
	 * 初始化获取未处理完的数据，放入队列。 同时启动一个deamon线程从队列中获取数据处理
	 */
	private void init() {
		log.info("spring.profiles.active:" + activeProfile);
		//只有线上环境，在项目启动时处理未处理完任务【原因：测试环境和线上环境使用同一个数据库】
		if("prod".equalsIgnoreCase(activeProfile)) {
			LambdaQueryWrapper<UserGroupEntity> queryWrapper = Wrappers.<UserGroupEntity>lambdaQuery();
			//只要不是成功，都放入到队列中，等待继续处理
			queryWrapper.ne(UserGroupEntity::getProcessStatus, UserGroupProcessStatus.success).orderByDesc(UserGroupEntity::getGroupId);
			List<UserGroupEntity> dataList = userGroupService.list(queryWrapper);
			if(dataList != null && !dataList.isEmpty()) {
				dataList.stream().forEach(item -> execute(item.getGroupId()));
			}
		}
		
		//启动一个守护进程开始跑任务
		//由于clickhouse性能考虑，一次只能执行一个任务
		Thread thread = new Thread("UserGroupCollectThread") {
			@Override
			public void run() {
				log.info("UserGroupCollectThread Deamon Thread is Starting...");
				while(true) {
					String groupId = null;
					try {
						//队列无数据时，阻塞线程
						groupId = queue.take();
						log.info("获取一个待处理的人群,开始处理|groupid={}", groupId);
						_deal(groupId);
					} catch (Exception e) {
						log.error(e.getMessage(), e);
						
						//遇到错误，设置人群数据为失败
						UserGroupEntity entity = new UserGroupEntity();
						entity.setGroupId(groupId);
						entity.setProcessStatus(UserGroupProcessStatus.failure);
						userGroupService.updateById(entity);
					}
				}
			}
		};
		thread.setDaemon(true);
		thread.start();
	}
	
	/**
	 * 数据处理入口
	 */
	private void _deal(String groupId) {
		//1.更新为正在处理状态
		UserGroupEntity entity = new UserGroupEntity();
		entity.setGroupId(groupId);
		entity.setProcessStatus(UserGroupProcessStatus.processing);
		userGroupService.updateById(entity);
		
		//计算出当前人群需要处理哪些月份的数据
		List<String> watingMonthList = Optional.ofNullable(_calcHandleMonthList(groupId)).orElse(new ArrayList<>());
		log.info("需要处理的月份列表|groupId={},theMonth={}", groupId, String.join(",", watingMonthList));
		//计算时间
		StopWatch stopWatch = new StopWatch("标签处理时间");
		stopWatch.start("处理用户ID关联表数据");
		//2.处理用户ID关联表数据
		long count = _dealUserIdJoinData(groupId);
		stopWatch.stop();
		log.info("groupUserId:finish|groupId={},cost={}s", groupId, stopWatch.getLastTaskTimeMillis() / 1000.f);
		if(count > 0) {
			//3.检查汇总表的标签列是否有变动
			stopWatch.start("检查汇总表的标签列是否有变动");
			List<AdsUserprofileTag> tagList = _checkTableColumn();
			stopWatch.stop();
			log.info("checkColumn:finish|groupId={},cost={}ms",groupId, stopWatch.getLastTaskTimeMillis());
			
			//4.汇总并入库
			for(String theMonth : watingMonthList) {
				log.info("start:calculate|groupid={}, month={}", groupId, theMonth);
				stopWatch.start("执行["+theMonth+"]月数据入库");
				//开始执行分批月份数据入库
				boolean dealRet = _dealCollectAndInsert(groupId, theMonth, tagList);
				stopWatch.stop();
				log.info("end:calculate|groupid={},month={},cost={}s", groupId, theMonth, stopWatch.getLastTaskTimeMillis() / 1000.f);
				if(dealRet) {
					//5.记录执行的日志
					userGroupAsyncDataLogService.save(new UserGroupAsyncDataLog(groupId, theMonth, 1, new Date()));
				}
			}
		}
		//6.更新人群表的状态为已完成
		UserGroupEntity tempEntity = new UserGroupEntity();
		tempEntity.setGroupId(groupId);
		tempEntity.setProcessStatus(UserGroupProcessStatus.success);
		tempEntity.setFinishTime(new Date());
		userGroupService.updateById(tempEntity);
		log.info("userGroup:end|groupId={},cost={}s",groupId, stopWatch.getTotalTimeSeconds());
		log.info(stopWatch.prettyPrint());
	}
	
	/**
	 * 处理用户ID关联表插入数据
	 * @param groupId
	 */
	private long _dealUserIdJoinData(String groupId) {
		//1.判断是否有groupId的数据
		long cnt = userGroupUserIdService.countByGroupId(groupId);
		if(cnt <= 0) {
			log.info("start:insertUserId|groupid={}", groupId);
			long startTime = System.currentTimeMillis();
			//2.查询人群构造查询条件
			UserGroupEntity userGroupEntity = userGroupService.getById(groupId);
			if(userGroupEntity == null) {
				return 0;
			}
			QueryWrapper<ZUserTagEntity> queryWrapper = userGroupRuleService.parse(userGroupEntity);
			//3.入库数据
			if(StrUtil.isBlank(userGroupEntity.getImportId())) {
				userGroupUserIdService.insertData(groupId, queryWrapper);
			}else {
				userGroupUserIdService.insertDataWithImport(groupId, userGroupEntity.getImportId(), queryWrapper);
			}
			cnt = userGroupUserIdService.countByGroupId(groupId);
			log.info("end:insertUserId|groupid={},count={},cost={}s", groupId, cnt, (System.currentTimeMillis() - startTime) / 1000.0f);
			//4.更新人群库的数量
			UserGroupEntity entity = new UserGroupEntity();
			entity.setGroupId(groupId);
			entity.setReachCount(String.valueOf(cnt));
			userGroupService.updateById(entity);
		}
		return cnt;
	}
	
	/**
	 * 校验表的列【删除/新增】
	 * 
	 * @throws SQLException
	 */
	private List<AdsUserprofileTag> _checkTableColumn() {
		//1. 查询clickhouse汇总表的字段
		List<JSONObject> dataList = userGroupSumDataService.getColumnList();
		List<String> ckColumnList = new ArrayList<>(dataList.size());
		for(JSONObject jsonObject : dataList) {
			String columnName = jsonObject.getString("name");
			if(!columnName.equalsIgnoreCase("group_id") && !columnName.equalsIgnoreCase("the_month")) {
				ckColumnList.add(columnName);
			}
		}
		//2. 查询tag字典表
		List<AdsUserprofileTag> tagList = userProfileService.list(Wrappers.<AdsUserprofileTag>lambdaQuery().eq(AdsUserprofileTag::getIsOnline, "1"));
		List<String> dictColumnList = new ArrayList<>(tagList.size());
		for(AdsUserprofileTag item : tagList) {
			dictColumnList.add(item.getEnglishName() + "_" + item.getTagId());
		}
		
		//3. 查询增加列和删除列
		List<String> addColumnList = dictColumnList.stream().filter(f -> !ckColumnList.contains(f)).collect(Collectors.toList());
		List<String> delColumnList = ckColumnList.stream().filter(f -> !dictColumnList.contains(f)).collect(Collectors.toList());
		
		//4. 增加列操作
		if(!addColumnList.isEmpty()) {
			StringBuilder sqlBuilder = new StringBuilder();
			sqlBuilder.append("alter table ads.BIZ_USER_GROUP_TAG_SUM ");
			for(String column : addColumnList) {
				sqlBuilder.append("add column ").append(column + " UInt32,");
			}
			sqlBuilder.deleteCharAt(sqlBuilder.length() - 1);
			log.info("start:addColumn|{}", String.join(",", addColumnList));
			userGroupSumDataService.addOrDelColumn(sqlBuilder.toString());
			userGroupSumDataService.cleanColumnListCache();
		}
		
		//5. 删除列操作
		if(!delColumnList.isEmpty()) {
			StringBuilder sqlBuilder = new StringBuilder();
			sqlBuilder.append("alter table ads.BIZ_USER_GROUP_TAG_SUM ");
			for(String column : delColumnList) {
				sqlBuilder.append("drop column ").append(column + ",");
			}
			sqlBuilder.deleteCharAt(sqlBuilder.length() - 1);
			log.info("start:delColumn|{}", String.join(",", delColumnList));
			userGroupSumDataService.addOrDelColumn(sqlBuilder.toString());
			userGroupSumDataService.cleanColumnListCache();
		}
		return tagList;
	}
	
	
	/**
	 * 汇总并入库
	 * @param groupId
	 * @param startMonth
	 */
	private boolean _dealCollectAndInsert(String groupId, String month, List<AdsUserprofileTag> tagList) {
		//先查询当前月是否存在数据
		long count = userGroupSumDataService.count(groupId, month);
		if(count > 0) {
			log.info("查找到数据，无需再次处理|groupId={},month={}", groupId, month);
			return true;
		}
		StringBuilder sqlBuilder = new StringBuilder();
		sqlBuilder.append("insert into ads.BIZ_USER_GROUP_TAG_SUM(");
		for (AdsUserprofileTag item : tagList) {
			sqlBuilder.append(item.getEnglishName() + "_" + item.getTagId() + ",");
		}
		sqlBuilder.append("group_id, the_month)\n");
		sqlBuilder.append("select ");
		for (AdsUserprofileTag item : tagList) {
			if ("array".equalsIgnoreCase(item.getTagCType())) {
				sqlBuilder.append("uniqIf(USERID_V,has(" + item.getEnglishName().toUpperCase() + ".E, '" + item.getTagId()
						+ "')),").append("\n");
			} else if ("EVENT_UNSPNAME".equalsIgnoreCase(item.getEnglishName())) {
				sqlBuilder.append("uniqIf(USERID_V,EVENT_SPNAME != '" + item.getTagId() + "'),").append("\n");
			} else {
				sqlBuilder.append("uniqIf(USERID_V," + item.getEnglishName().toUpperCase() + " = '" + item.getTagId()
						+ "'),").append("\n");
			}
		}
		sqlBuilder.append("'" + groupId + "', '"+month+"'").append("\n");
		sqlBuilder.append("from Z_USER_TAG_FLAT_ALL ua").append("\n")
				.append("global inner join ads.BIZ_USER_GROUP_USERID tum on ua.USERID_V = tum.user_id").append("\n")
				.append("where tum.user_group_id = '"+groupId+"' and THEMONTH_V = '"+month+"'");
		userGroupSumDataService.batchSaveData(sqlBuilder.toString());
		return true;
	}
	
	/**
	 * 计算当前人群需要处理哪些月份数据
	 * @param groupId
	 * @return
	 */
	private List<String> _calcHandleMonthList(String groupId){
		UserGroupEntity userGroupEntity = userGroupService.getById(groupId);
		if(userGroupEntity == null) {
			return null;
		}
		//待处理的日期列表
		List<String> waitingTheMonthList = JSON.parseArray(userGroupEntity.getAnalyseDateJson(), String.class);
		//已处理的日期列表
		LambdaQueryWrapper<UserGroupAsyncDataLog> _queryWrapper = Wrappers.<UserGroupAsyncDataLog>lambdaQuery().select(UserGroupAsyncDataLog::getTheMonth).eq(UserGroupAsyncDataLog::getGroupId, groupId);
		List<String> alreayTheMonthList = userGroupAsyncDataLogService.listObjs(_queryWrapper, t -> t.toString());
		//计算得到需要处理的日期列表
		List<String> needTheMonthList = null;
		if(waitingTheMonthList != null && !waitingTheMonthList.isEmpty()) {
			needTheMonthList = waitingTheMonthList.stream().filter(t -> alreayTheMonthList == null || !alreayTheMonthList.contains(t)).sorted().collect(Collectors.toList());
		}
		return needTheMonthList;
	}

	
	@Override
	public void run(String... args) throws Exception {
		if(flag) {
			init();
		}
	}
}
